/*-
 * NEC Mobile Gear APM (Advanced Power Management) 
 *
 * Copyright (c) 1998 Shin Takemura all rights reserved.
 * Copyright (c) 1998 Kenji Ohta all rights reserved.
 * Copyright (c) 1998 NAKAGAWA, Yoshihisa all rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote 
 *    products derived from this software without specific prior written 
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: necmg_apm.c,v 1.1.4.3 1998/12/28 04:07:16 nakagawa Exp $
 *
 */
#ifdef NECMG
#include <sys/param.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <machine/md_var.h>
#include <machine/cpufunc.h>
#include <machine/clock.h>

#include <vm/vm.h>
#include <vm/vm_param.h>
#include <vm/pmap.h>

#include <machine/apm_bios.h>
#include <i386/apm/apm_setup.h>
#include <necmg/necmg/necmg.h>
#include <opt_necmg.h>

#include "card.h"

static void necmg_apm_timeout __P((void *dummy));
static int check_power_keys __P((void));
static void check_power_status __P((void));
static void register_disp __P((void));

static int set_serial_power(int on);
static int set_modem_power(int on);

static int mg_suspend __P((void));
static void in22(unsigned char idx, unsigned char *data);
static void out22(unsigned char idx, unsigned char data);
static void in28(unsigned char idx, unsigned char *data);
static void out28(unsigned char idx, unsigned char data);
static int waitfor(unsigned char mask, unsigned short port,
		   int flag, int timeout);
static int write_readF18(unsigned short cmd, unsigned short *data);
static int readF18(unsigned char *data, int n);
static int writeF18(unsigned char *data, int n);
static int outF18(unsigned char data);
static int inF18(unsigned char *data);

#define TIMEOUT 0x4000
#define TIMER_INTERVAL	(hz / 20)  /* about 50 msec */
#define MSEC	1000
#define NECMG_BATT_UNKNOWN	255
#define NECMG_BATT_TYPE1	1
#define NECMG_BATT_TYPE2	2
#define NECMG_ACLINE_UNKNOWN	255
#define NECMG_ACLINE_OFFLINE	0
#define NECMG_ACLINE_ONLINE	1
#define NECMG_BATTLEVEL_UNKNOWN	255
#define NECMG_BATTSTAT_UNKNOWN	255
#define NECMG_BATTSTAT_HIGH	0
#define NECMG_BATTSTAT_LOW	1
#define NECMG_BATTSTAT_CRITICAL 2
#define NECMG_BATTLIFE_UNKNOWN	255

#define NECMG_BATTLEVEL_NICONS 6
#define NECMG_ACONLINE_ICON 1
#define NECMG_BATTLEVEL_ICON 2

/*
 *  These values are experimental.
 */
#define NECMG_BATTLEVEL_FULL		0xe8
#define NECMG_BATTLEVEL_HIGH		0xd4
#define NECMG_BATTLEVEL_LOW		0xc3
#define NECMG_BATTLEVEL_CRITICAL	0xc0

static unsigned char ac_line = NECMG_ACLINE_UNKNOWN;
static int batt_type = NECMG_BATT_UNKNOWN;
static unsigned char batt_level = NECMG_BATTLEVEL_UNKNOWN;
static unsigned char batt_stat = NECMG_BATTSTAT_UNKNOWN;
static int blink = 0;

/*
 *  global variables (Thsy comes from apm_setup.s)
 */
u_long	apm_version;
u_long	apm_cs_entry;
u_short	apm_cs32_base;
u_short	apm_cs16_base;
u_short	apm_ds_base;
u_short	apm_cs32_limit;
u_short	apm_cs16_limit;
u_short	apm_ds_limit;
u_short	apm_flags;
u_short	kernelbase;

/*
#define NECMG_APM_DEBUG
*/

/* static data */
struct necmg_apm_softc {
  int dummy;
};
static struct necmg_apm_softc necmg_apm_softc;
static initialized = 0;
static int event_suspend = 0;
static int event_battlow = 0;
static int serial_power = 1; /* When system boot up
				COM1 power is already ON. */
static int modem_power = 0;

void
necmg_apm_init()
{
  if (initialized)
    return;

  necmg_apm_timeout(NULL);

  initialized = 1;
  return;
}

static void
necmg_apm_timeout(void *dummy)
{
  struct necmg_apm_softc *sc = &necmg_apm_softc;
  static int tick = 0;

#ifdef NECMG_APM_DEBUG
  register_disp();
#endif
  check_power_keys();
  if ((tick++ % (hz/TIMER_INTERVAL)) == 0) {
    /* Once per second, may be. */
    blink++;
    check_power_status();
  }

  timeout(necmg_apm_timeout, NULL, TIMER_INTERVAL);
}

/*
 *  Fake BIOS entry point.
 */
int 
necmg_apm_int(u_long *eax, u_long *ebx, u_long *ecx)
{
  static int mg_resumed = 0;

  if (*eax == ((APM_BIOS << 8) | APM_SETPWSTATE) &&
      *ebx == PMDV_ALLDEV && *ecx == PMST_SUSPEND) {
    /*
     *  SUSPEND
     */ 
#ifdef NECMG_APM_DEBUG
    printf("fake APM BIOS: go suspend\n");
#endif
    mg_suspend();
    mg_resumed = 1;
    return 0;
  } else
  if (*eax == ((APM_BIOS << 8) | APM_GETPWSTATUS) &&
      *ebx == PMDV_ALLDEV && *ecx == 0) {
    /*
     *  GET POWER STATUS
     */ 
    *ebx = (ac_line << 8) | batt_stat;
    *ecx = NECMG_BATTLIFE_UNKNOWN;
/*#ifdef NECMG_APM_DEBUG
*/
#if 1
    printf("necmg power status: type=%d   level=%02x\n",
	    batt_type, batt_level);
#endif
    return 0;
  } else
  if (*eax == ((APM_BIOS << 8) | APM_GETPMEVENT)) {
    /*
     *  GET PM EVENT
     */ 
    if (mg_resumed) {
      *ebx = PMEV_NORMRESUME;
      mg_resumed = 0;
    } else
    if (event_suspend) {
      *ebx = PMEV_USERSUSPENDREQ;
      event_suspend = 0;
    } else
    if (event_battlow) {
      *ebx = PMEV_BATTERYLOW;
      event_battlow = 0;
    } else {
      *ebx = PMEV_NOEVENT;
    }
    return 0;
  }
  return 1;
}

/*
 * Power off
 */
void
necmg_cpu_poweroff(int howto, void *dummy_arg) {
  printf("\nGood bye!\n");
  /* wait 1 sec for printf's to complete and be read */
  DELAY(1000000);

  /*
   *  I don't know why this cause system power off.
   *  But it works fine on MC-MK12. (takemura)
   */

  /* force a shutdown by unmapping entire address space ! */
  bzero((caddr_t) PTD, PAGE_SIZE);

  /* "good night, sweet prince .... <THUNK!>" */
  invltlb();
  /* NOTREACHED */
  while(1);
}

int
necmg_set_contrast(int high_low)
{
  int s, res;
  unsigned char cmd[2][8] = {
    { 0x70, 0x76, 0x70, 0x74, 0x70, 0x76, 0x70, 0x77 },
    { 0x70, 0x72, 0x70, 0x70, 0x70, 0x72, 0x70, 0x77 }
  };

  s = splimp();
  outb(0xf11, 0x02);
  outb(0xf11, 0x00);
  outb(0xf11, 0x80);

  res = writeF18(cmd[high_low ? 1 : 0], 8);

  outb(0xf11, 0x02);
  outb(0xf11, 0x00);
  splx(s);

  return (res);
}

static void
register_disp()
{
  int x = 100;
  unsigned char d0, d1, d2;
  static unsigned db0 = 0, db1 = 0, db2 = 0;
  char *p, buf[100];

  d0 = inb(0xf12);
  d1 = inb(0xf15);
  d2 = inb(0xf18);

  if (d0 == db0 && d1 == db1 && d2 == db2) {
    return;
  }
  sprintf(buf, "%02x %02x %02x", d0, d1, d2);

  db0 = d0;
  db1 = d1;
  db2 = d2;
  for (p = buf; *p; p++) {
    necmg_draw_char(x, 0, *p & 0xff, 7, 8);
    x += 2;
  }
}

/*-----------------------------------------------------------------------------

  COM port power control

-----------------------------------------------------------------------------*/
int
necmg_com_power(int port, int on)
{
#ifdef NECMG_APM_DEBUG
    printf("COM%d %s\n", port, on ? "on" : "off");
#endif
  switch(port) {
  case NECMG_COM1:
    serial_power = on;
    return (set_serial_power(on));
    break;
  case NECMG_COM3:
    modem_power = on;
    return (set_modem_power(on));
  default:
    panic("necmg_com_power(): invalied port");
  }
  return (0);
}

static int
set_serial_power(int on)
{
  disable_intr();
/*
  if (res = COM1Power_check_bx(bx)) {
    return (res);
  }
*/
  if (on) {
    /* Power ON */
    outb(0xf02, inb(0xf02) | 0x80);
    outb(0xf15, inb(0xf15) & ~0x01);
    DELAY(120 * MSEC);
    outb(0xf12, inb(0xf12) & ~0x08); /* COM1 on */
    DELAY(70 * MSEC);
    outb(0xf08, (inb(0xf08) & 0xfa) | 0x2);
  } else {
    /* Power OFF */
    outb(0xf08, inb(0xf08) & 0xf8);
    outb(0xf12, inb(0xf12) | 0x08); /* COM1 off */
    /* XXX, This makes system hungup.
    outb(0xf15, inb(0xf15) | 0x01);
    */
    DELAY(85 * MSEC);
  }
  enable_intr();
  return (0);
}

static int
set_modem_power(int on)
{
  int sound;
  disable_intr();
  if (on) {
    if ((inb(0xf12) & 0x10) == 0) return (0);
    outb(0xf14, inb(0xf14) & 0xdf);

    sound = !(inb(0xf12) & 0x4);        /* Sound on/off ? */
    outb(0xf12, (inb(0xf12) & ~0x20) | 0x10 | 0x4);     /* modem, sound off */
    DELAY(10 * MSEC);

    outb(0xf12, inb(0xf12) & 0xef);     /* Modem on */
    DELAY(200 * MSEC);

    outb(0xf12, inb(0xf12) | 0x20);     /* 0x20 ??? off ? */
    DELAY(10 * MSEC);

    /* Sound ON/OFF restore */
    if (sound) {
      outb(0xf12, inb(0xf12) & ~0x04);  /* sound ON */
    }

    /*
     *  very very long wait.
     */
    /* XXX, I can't stand this.
       DELAY(2500 * MSEC);
       */
    DELAY(2000 * MSEC);
  } else {
    sound = !(inb(0xf12) & 0x04);
    outb(0xf12, inb(0xf12) | 0x04);             /* sound off */

    outb(0xf12, inb(0xf12) & ~0x20);    /* ??? on ?*/
    DELAY(10 * MSEC);

    if (sound) {
      outb(0xf12, inb(0xf12) & ~0x04);  /* sound on */
    }
    outb(0xf12, inb(0xf12) | 0x10);             /* modem off */
  }
  enable_intr();

  return (0);
}

/*-----------------------------------------------------------------------------

  POWER STATUS

-----------------------------------------------------------------------------*/
static void
check_power_status()
{
  unsigned char tmpb;
  unsigned short tmpw;

  unsigned char level_table[][NECMG_BATTLEVEL_NICONS + 1] = {
    /* fatal 0    1     2     3     4     5 */
    {  0xc3, 0xc5, 0xcb, 0xd3, 0xdc, 0xe7, 0xf4 }, /* dry battery */
						   /* 0xc3+pow(i,1.6)*2.8 */
    {  0xc0, 0xc3, 0xc7, 0xcd, 0xd4, 0xdc, 0xe4 }, /* cargable battery */
						   /* 0xc0+pow(i,1.4)*3 */
  };

  int pos = 77;
  int icon = -1, color;
  static int prev_icon = -1;
  static int prev_color = -1;

  /*
   *  AC or battery
   */
  in28(0x0d, &tmpb);
  if (tmpb & 0x10) {
    ac_line = 1;
  } else {
    ac_line = 0;
  }

  /*
   *  Dry battery or rechargable
   */
  if (write_readF18(0x4030, &tmpw) != 0 ||
      (tmpw & 0xc0f0) != 0xc0f0) {
    /* error */
    printf("necmg: write_readF18(battery type) failed.\n");
  } else {
    if (tmpw & 0x0008) {
      batt_type = NECMG_BATT_TYPE1;
    } else {
      batt_type = NECMG_BATT_TYPE2;
    }
  }

  /*
   *  Battery level
   */
  if (write_readF18(0x4050, &tmpw) != 0 ||
      (tmpw & 0xc0f0) != 0xc0d0) {
    /* error */
    printf("necmg: write_readF18(battery level) failed.\n");
    batt_level = NECMG_BATTLEVEL_UNKNOWN;
  } else {
    batt_level = ((tmpw >> 8) & 0x3f) | ((tmpw<<6) & 0xc0);
    if (0xff <= batt_level) {
      batt_level = 0xfe;
    }
  }

  if (batt_level == NECMG_BATTLEVEL_UNKNOWN) {
    icon = NECMG_BATTLEVEL_ICON;
    color = NECMG_LIGHTGRAY;
    batt_stat = NECMG_BATTSTAT_UNKNOWN;
  } else {
    int level;
    unsigned char *table;
    if (batt_type == NECMG_BATT_TYPE1) {
      /* battery type is type1 */
      table = level_table[0];
    } else {
      /* battery type is type2 or unknown */
      table = level_table[1];
    }
    
    for (level = NECMG_BATTLEVEL_NICONS - 1; -2 < level; level--) {
      if (table[level + 1] <= batt_level) break;
    }
    
    if (level < 0) {
      icon = NECMG_BATTLEVEL_ICON + 0;
      color = (blink % 2) ? NECMG_BLACK : NECMG_WHITE;
      batt_stat = NECMG_BATTSTAT_CRITICAL;
      if (level < -1) {
	/*
	 *  DANGER !!
	 *  battery level is too low.
	 */
	event_battlow = 1;
      }
    } else {
      icon = NECMG_BATTLEVEL_ICON + level;
      color = NECMG_BLACK;
      batt_stat = (2 < level) ? NECMG_BATTSTAT_HIGH : NECMG_BATTSTAT_LOW;
    }
  }
  if (ac_line == NECMG_ACLINE_ONLINE) {
    /*
     * AC line status is online.
     */
    if (batt_stat == NECMG_BATTSTAT_HIGH || !(blink % 2)) {
      /*
       * and battery voltage is green.
       */
      icon = NECMG_ACONLINE_ICON;
      color = NECMG_BLACK;
    }
  }

  /*
   *  Draw power status icon.
   */
  if (icon != -1) {
    if (prev_icon != icon || prev_color != color || (blink % 20) == 0) {
      necmg_draw_icon(pos * 2, 0,
		      necmg_icon16x8[icon],
		      4, 8, color, NECMG_WHITE);
      prev_icon = icon;
      prev_color = color;
    }
  }

#ifdef NECMG_APM_DEBUG
  {
    int x = 4;
    char *p, buf[100];
    unsigned char tmp0, tmp1;

    sprintf(buf, "%s %d %02x %d",
	    ac_line ? "AC" : "  ",
	    batt_type,
	    batt_level,
	    batt_stat);
    for (p = buf; *p; p++) {
      necmg_draw_char(x, 0, *p & 0xff, 7, 8);
      x += 2;
    }
  }
#endif
}

/*-----------------------------------------------------------------------------

  POWER KEY

-----------------------------------------------------------------------------*/
static int
check_power_keys()
{
  int s, i;
  unsigned char tmp, data[2];

  s = splimp();
  if (inb(0xf12) & 0x02) {
    splx(s);
    return (0);
  }

  tmp = inb(0xf15);
  for (i = 0; i < 2; i++) {
    if (inF18(&data[i])) {
      /* timeout */
      outb(0xf15, tmp);
      splx(s);
      return (0);
    }
  }

  outb(0xf15, tmp);
  splx(s);

#ifdef NECMG_APM_DEBUG
  printf("check_power_key: %02x %02x\n", data[0], data[1]);
#endif

  if ((data[0] & 0xf0) == 0xa0 && (data[1] & 0xc0) == 0xc0) {
    if (data[1] & 0x01) {
      /* cover switch status changed */
      if ((data[0] & 0x01) == 0) {
	/* cover was closed */
	event_suspend = 1;
      }
    }
    if (data[1] & 0x02) {
      /* power switch status changed */
      if (data[0] & 0x02) {
	/* power key down */
	event_suspend = 1;
      } else {
	/* power key up */
      }
    }
  }

  return (0);
}

/*-----------------------------------------------------------------------------

  POWER MANAGEMENT PROCESSOR I/O ROUTINE

  I/O:
	F12h, F15h, F18h, F19h

-----------------------------------------------------------------------------*/
static int
write_readF18(unsigned short cmd, unsigned short *data)
{
  int s, res;
  int errflag;
  unsigned char data0, data1;
  unsigned char buf[2];

  s = splimp();
  outb(0xf11, 0x02);
  outb(0xf11, 0x00);
  outb(0xf11, 0x80);

  buf[0] = (cmd & 0xff);
  buf[1] = ((cmd >> 8) & 0xff);
  res = writeF18(buf, 2);
  if (res) goto exit_func;

  res = readF18(buf, 2);
  if (res) goto exit_func;

 exit_func:
  outb(0xf15, (inb(0xf15) & 0xf9) | 0x02);

  outb(0xf11, 0x02);
  outb(0xf11, 0x00);
  splx(s);

  *data = buf[1] * 256 + buf[0];

  return (res);
}

static int
readF18(unsigned char *data, int n)
{
  int i, res;
  int s;

  s = splimp();
  /*
   *  === Failsafe ===
   *  Usually F19h = 0xff and 0xF15 = -----01-
   *  so this block is not need.
   */
  outb(0xf19, 0xff);
  outb(0xf15, (inb(0xf15) & 0xf9) | 0x02);
  res = waitfor(0x40, 0xf15, 1, TIMEOUT);
  if (res) goto exit_func;

  for (i = 0; i < n; i++) {
    res = inF18(&data[i]);
    if (res) {
      goto exit_func;
    }
  }
  res = 0;

 exit_func:
  splx(s);
#ifdef NECMG_APM_DEBUG
  printf("readF18(%d bytes) = %d\n", n, res);
#endif

  return (res);
}

static int
writeF18(unsigned char *data, int n)
{
  int i, j, res;
  int s;

  s = splimp();

  outb(0xf19, 0x00);

  for (i = 0; i < n; i += 2) {
    outb(0xf15, (inb(0xf15) & 0xf9) | 0x00);
    for (j = 1000; 0 < j; j--) {
      DELAY(50);
      check_power_keys();
      if ((inb(0xf15) & 0x40) == 0) break;
    }
    if (j <= 0) goto exit_func;

    res = outF18(data[i]);
    if (res) goto exit_func;

    res = outF18(data[i + 1]);
    if (res) goto exit_func;

    outb(0xf15, (inb(0xf15) & 0xf9) | 0x02);
    res = waitfor(0x40, 0xf15, 1, TIMEOUT);
    if (res) goto exit_func;
  }

 exit_func:
  outb(0xf19, 0xff);
  outb(0xf15, (inb(0xf15) & 0xf9) | 0x02);
  splx(s);

#ifdef NECMG_APM_DEBUG
  printf("writeF18(%d bytes) = %d\n", n, res);
#endif

  return (res);
}

static int
outF18(unsigned char data)
{
  int res;

  outb(0xf18, data);
  outb(0xf15, (inb(0xf15) & 0xf9) | 0x04);
  res = waitfor(2, 0xf12, 0, TIMEOUT);
  if (res) goto exit_func;
  outb(0xf15, (inb(0xf15) & 0xf9) | 0x00);
  res = waitfor(0x02, 0xf12, 1, TIMEOUT);

 exit_func:
#ifdef NECMG_APM_DEBUG
  if (res) {
    printf("outF18(%02x): res=%d\n", data, res);
  }
#endif
  return (res);
}

static int
inF18(unsigned char *data)
{
  int res;
  outb(0xf15, (inb(0xf15) & 0xf9) | 0x02);
  res = waitfor(2, 0xf12, 0, TIMEOUT);
  if (res) goto exit_func;
  *data = inb(0xf18);
  outb(0xf15, (inb(0xf15) & 0xf9) | 0x06);
  res = waitfor(2, 0xf12, 1, TIMEOUT);

 exit_func:
#ifdef NECMG_APM_DEBUG
  if (res) {
    printf("inF18(): inb(0xf18)=%02x  res=%d\n", *data, res);
  }
#endif
  return (res);
}

static int
waitfor(unsigned char mask, unsigned short port, int flag, int timeout)
{
  int res, t;
  unsigned char data;
  static count = 0;

  count++;

  for (t = 0; t < timeout; t++) {
    DELAY(100);
    if (flag) {
      if ((data = inb(port)) & mask) {
	break;
      }
    } else {
      if (((data = inb(port)) & mask) == 0) {
	break;
      }
    }
  }
  if (timeout <= t) {
#ifdef NECMG_APM_DEBUG
    printf("%d: waitfor(%x, %x, %d, %x): rest=%x inb(%x)=%02x\n",
	    count, mask, port, flag, timeout,
	    timeout - t, port, data);
#endif
  }
  return (t < timeout ? 0 : 1);
}

/*-----------------------------------------------------------------------------

  SUSPEND / RESUME

  I/O:
	22h, 23h
	28h, 2Ah

-----------------------------------------------------------------------------*/
static int
mg_suspend(void)
{
  unsigned char tmp0, tmp1;

  disable_intr();
  in22(0x2e, &tmp0);
  in22(0x2f, &tmp0);
  out22(0x2e, 0x84);
  out22(0x2f, 0x50);

  out22(0x2e, tmp0);
  out22(0x2f, tmp1);
  
  in28(0x0a, &tmp0);
  out28(0x0a, 0x8f);

  if (serial_power) set_serial_power(0);
  if (modem_power) set_modem_power(0);

  /*
   * set start up info 0xb. 
   */
  outb(0x70, 0x8f);
  outb(0x71, 0x0b);

  /*
   * suspend
   */
  out22(0x01, 0x10);

#if NCARD == 0
  /*
   *  If we have no pccard service, make ATA flash happy here.
   */
  necmg_wd_pccard_enabler();
#endif /* NCARD */

  if (serial_power) set_serial_power(1);
  if (modem_power) set_modem_power(1);

  enable_intr();
  return 0;
}

static void
in22(unsigned char idx, unsigned char *data)
{
  unsigned char tmp;
  tmp = inb(0x22);
  outb(0x22, idx);
  *data = inb(0x23);
  outb(0x22, tmp);
}

static void
out22(unsigned char idx, unsigned char data)
{
  unsigned char tmp;
  tmp = inb(0x22);
  outb(0x22, idx);
  outb(0x23, data);
  outb(0x22, tmp);
}

static void
in28(unsigned char idx, unsigned char *data)
{
  unsigned char tmp;
  tmp = inb(0x28);
  outb(0x28, idx);
  *data = inb(0x2a);
  outb(0x28, tmp);
}

static void
out28(unsigned char idx, unsigned char data)
{
  unsigned char tmp;

  tmp = inb(0x28);
  outb(0x28, idx);
  outb(0x2a, data);
  outb(0x28, tmp);
}

#endif /* NECMG */
