/*-
 *
 * Copyright (c) 1998 Shin Takemura all rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote 
 *    products derived from this software without specific prior written 
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: necmg_cons.c,v 1.1.4.1 1998/12/05 22:49:20 itojun Exp $
 *
 */
#ifdef NECMG
#include <sys/param.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <machine/md_var.h>
#include <machine/cpufunc.h>
#include <machine/clock.h>

#include <vm/vm.h>
#include <vm/vm_param.h>
#include <vm/pmap.h>

#include <necmg/necmg/necmg.h>
#include <opt_necmg.h>

unsigned char *necmg_fontrom_base = 0;
static mg_zoompos = -1;

void
necmg_draw_icon(int x, int y, unsigned char *icon, int width, int height,
	     unsigned char fore,  unsigned char back)
{
  unsigned char *fb = NECMG_FRAMEBUFFER;

  fb += x + y * 0x100;

  for (y = 0; y < height; y++) {
    for (x = 0; x < width; x++) {
      *fb++ = (fore & *icon) | (back & ~*icon);
      icon++;
    }    
    fb += (0x100 - width);
  }
}

static unsigned char *
necmg_font_data(int code)
{
  static unsigned char null_font8[] = {
    0xfe, 0x82, 0x82, 0x82,
    0x82, 0x82, 0x82, 0x82,
    0x82, 0x82, 0x82, 0x82,
    0x82, 0x82, 0xfe, 0x00,
  };
  static unsigned char null_font16[] = {
    0xff, 0xfe, 0x80, 0x02, 0x80, 0x02, 0x80, 0x02, 0x80, 0x02, 0x80, 0x02,
    0x80, 0x02, 0x80, 0x02, 0x80, 0x02, 0x80, 0x02, 0x80, 0x02, 0x80, 0x02,
    0x80, 0x02, 0x80, 0x02, 0xff, 0xfe, 0x00, 0x00
  };

  if (code < 0x100) {
    /* 1 byte code */
    if (necmg_fontrom_base == 0)
      (null_font8); /* kanji ROM not available yet */
    return (necmg_fontrom_base + 0x83800 + code * 16);
  }

  /* 2 byte code */
  if (necmg_fontrom_base == 0) (null_font16); /* kanji ROM not available yet */
  return necmg_fontrom_base +
    (code & 0x7f00) * 0x0c +
      (code & 0x7f) * 0x20 + 0x2b800;
}

static void
necmg_put_char(int x, int y, int code, int attr)
{
  int zoom = (y == mg_zoompos && 0 <= mg_zoompos);

  x = x * 2;
  if (mg_zoompos < y && 0 <= mg_zoompos) {
    y = y * 9 + 8 + 7;
  } else {
    y = y * 9 + 8;
  }
  necmg_draw_char(x, y, code, attr, zoom ? 16 : 9);
}

/* EUC kanji code
 * 2 byte code (variable length code)
 */
#define iskanji1(c) (0xa0 <= (unsigned)(c) && (unsigned)(c) < 0xff)

void
necmg_cons_update(int xsize, int ysize, 
		  unsigned short *vvram, int start, int end)
{
  int i, flag = 0;
  int code, attr;

  if (start % xsize) {
    for (i = (start / xsize) * xsize; i < start; i++) {
      if (iskanji1(vvram[i])) {
	flag = 1;
	i++;
      } else {
	flag = 0;
      }
    }
  }
  if (flag) {
    start -= 1;
  }

  for (i = start; i <= end; i++) {
    int x = i % xsize;
    int y = i / xsize + 0;
    if (ysize <= y) break;
    if (iskanji1(vvram[i]&0xff) && x < xsize - 1) {
      /* 2 byte code */
      attr = (vvram[i] >> 8) & 0xff;
      code = (vvram[i++] & 0xff) << 8;
      code |= (vvram[i] & 0xff);
    } else {
      code = vvram[i] & 0xff;
      attr = (vvram[i] >> 8) & 0xff;
    }
    necmg_put_char(x, y, code, attr);
  }
}

void
necmg_cons_clear()
{
  unsigned char *fb = NECMG_FRAMEBUFFER;

  int y;
  for (y = 0; y < 240; y++) {
    bzero(fb, 640/4);
    fb += 0x100;
  }
}

#ifdef NECMG_ZOOMCONS
int
necmg_cons_get_zoompos()
{
  return (mg_zoompos);
}

int
necmg_cons_set_zoompos(int xsize, int ysize, unsigned short *vvram, int y)
{
  int start, end;
  int res = mg_zoompos;

  if (ysize <= y) {
    y = ysize - 1;
  }
  if (mg_zoompos == y) {
    return (mg_zoompos);
  }

  if (mg_zoompos < 0) {
    if (y < 0) {
      /* nothing to do */
      return (res);
    } else {
      mg_zoompos = y;
      start = y;
      end = ysize - 1;
    }
  } else {
    if (y < 0) {
      start = mg_zoompos;
      end = ysize - 1;
      mg_zoompos = -1;
      { /* clear bottom line */
	int i;
	for (i = 0; i < xsize; i++) {
	  necmg_draw_char(i * 2, ysize * 9 + 8, ' ', 0, 9);
	}
      }
    } else {
      if (mg_zoompos < y) {
	start = mg_zoompos;
	end = y;
	mg_zoompos = y;
      } else {
	start = y;
	end = mg_zoompos;
	mg_zoompos = y;
      }
    }
  }

  necmg_cons_update(xsize, ysize, vvram, start * xsize, (end + 1) * xsize - 1);

  return (res);
}
#endif

void
necmg_draw_char(int x, int y, int code, int attr, int size)
{
  unsigned char *fb = NECMG_FRAMEBUFFER;
  unsigned char *p, fore, back;
  struct necmg_font_s f;
  static data_table0[] = {
    0x00, 0x40, 0x10, 0x50, 0x04, 0x44, 0x14, 0x54,
    0x01, 0x41, 0x11, 0x51, 0x05, 0x45, 0x15, 0x55
  };
  static data_table1[] = {
    0x00, 0xc0, 0x30, 0xf0, 0x0c, 0xcc, 0x3c, 0xfc,
    0x03, 0xc3, 0x33, 0xf3, 0x0f, 0xcf, 0x3f, 0xff
  };
  static color_table[] = {
    /* 000 */ NECMG_WHITE,
    /* 001 */ NECMG_LIGHTGRAY,
    /* 010 */ NECMG_LIGHTGRAY,
    /* 011 */ NECMG_DARKGRAY,
    /* 100 */ NECMG_LIGHTGRAY,
    /* 101 */ NECMG_DARKGRAY,
    /* 110 */ NECMG_DARKGRAY,
    /* 111 */ NECMG_BLACK
  };

  f = necmg_fonts[(attr & 0x08) ? 1 : 0]; /* 0x08 bit is Foreground Intensity*/
  fore = color_table[attr & 0x7];
  back = color_table[(attr >> 4) & 0x7];
  fb += x + y * 0x100;

  if (0x100 <= code) {
    /* 16bit code */
    unsigned char *font_data = necmg_font_data(code);
    if (size == 16) {
      for (y = 0; y < 16; y++) {
	unsigned char d;
	d = data_table1[(*font_data>>4)&0x0f];
	*fb++ = (fore & d) | (back & ~d);
	d = data_table1[*font_data&0x0f];
	*fb++ = (fore & d) | (back & ~d);
	font_data++;
	d = data_table1[(*font_data>>4)&0x0f];
	*fb++ = (fore & d) | (back & ~d);
	d = data_table1[*font_data&0x0f];
	*fb++ = (fore & d) | (back & ~d);
	font_data++;
	fb += (0x100 - 4);
      }
    } else {
      for (y = 0; y < 8; y++) {
	unsigned char d, d0, d1;
	d0 = data_table0[(*font_data>>4)&0x0f];
	d1 = data_table0[(*(font_data+2)>>4)&0x0f];
	d = ((d0 | d1) << 1) | (d0 & d1);
	*fb++ = (fore & d) | (back & ~d);
	d0 = data_table0[*font_data&0x0f];
	d1 = data_table0[*(font_data+2)&0x0f];
	d = ((d0 | d1) << 1) | (d0 & d1);
	*fb++ = (fore & d) | (back & ~d);
	font_data++;

	d0 = data_table0[(*font_data>>4)&0x0f];
	d1 = data_table0[(*(font_data+2)>>4)&0x0f];
	d = ((d0 | d1) << 1) | (d0 & d1);
	*fb++ = (fore & d) | (back & ~d);
	d0 = data_table0[*font_data&0x0f];
	d1 = data_table0[*(font_data+2)&0x0f];
	d = ((d0 | d1) << 1) | (d0 & d1);
	*fb++ = (fore & d) | (back & ~d);

	font_data += 3;
	fb += (0x100 - 4);
      }
      if (size == 9) {
	bzero(fb, 4);
      }
    }
  } else {
    if (size == 16) {
      unsigned char d;
      p = necmg_font_data(code);
      for (y = 0; y < 16; y++) {
	d = data_table1[(*p>>4)&0x0f];
	*fb++ = (fore & d) | (back & ~d);
	d = data_table1[(*p>>0)&0x0f];
	*fb++ = (fore & d) | (back & ~d);
	p++;
	fb += (0x100 - 2);
      }
    } else {
      if (code < 0 || f.nchars <= code) {
	code = 0;
      }
      p = f.data + f.width * f.height * code;
      for (y = 0; y < f.height; y++) {
	for (x = 0; x < f.width; x++) {
	  *fb++ = (fore & *p) | (back & ~*p);
	  p++;
	}    
	fb += (0x100 - f.width);
      }
      if (size == 9) {
	bzero(fb, f.width);
      }
    }
  }
}

#endif /* NECMG */
